import numpy as np
import scipy as sp
import os
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from scipy.stats import binned_statistic
#from kapteyn import kmpfit
#import seaborn as sns

#from Binning_scipy import bin_data_given_bin_center
from glob import glob1

def bin_edge2_bin_center(bin_edges):
    center = np.zeros(len(bin_edges)-1)
    for i in range(len(bin_edges)-1):
        center[i] = bin_edges[i] + (bin_edges[i+1]-bin_edges[i])/2
    return center

def bin_center2_bin_edge(bin_center):
    edges = np.zeros(len(bin_center)+1)
    bin_sep = (bin_center[1]-bin_center[0])/2  # assume fixed bin width
    for i in range(len(bin_center)):
        #         print (i,bin_center[i])
        edges[i] = bin_center[i] - bin_sep
    edges[-1] = bin_center[-1] + (bin_center[-1]-bin_center[-2])/2
    return edges


def bin_data_given_bin_center(x, y, bin_center=[], nbin=10, Range=None, opt='mean'):
    '''
    Bin data, specifying the bin center
    arg :
    1. x : xdata
    2. y : ydata
    3. bin_center = an array of bin_center (default is None)
    4. nbin = 10 if the bin_center is None
    5. Range = None , given in the form (min(x),max(x))
    6. opt = sum (for statistics)
    count : compute the count of points within each bin. This is identical to an unweighted histogram. values array is not referenced.
    sum : compute the sum of values for points within each bin. This is identical to a weighted histogram.
    mean: compute the mean of the y-values in the same bin: useful for weighted data.
    std: compute the error bar
    Return:
    1. bin_center
    2. bin_counter
    '''

    if len(bin_center) != 0:
        bin_edges = bin_center2_bin_edge(bin_center)
    else:
        bin_edges = np.linspace(min(x), max(x), nbin)

    ret = binned_statistic(x, y, statistic=opt, bins=bin_edges, range=Range)
    return bin_edge2_bin_center(ret.bin_edges), ret.statistic

directory = r'C:/Users/pc2009/Dropbox (Heriot-Watt University Team)/RES_EPS_Quantum_Photonics_Lab/Experiments/Current Experiments/SiC_Vanadium/g2/PL_Map_DLucky2b/'
filename_x = os.path.join(directory, 'g2_DLucky_2b_Laser_Stable_234_42457THz_2_5uW_Bin_5ns_2022-05-17_09-23-04.txt')
x = np.loadtxt(filename_x)

#Create two arrays of given shape (len(x)) and type, with ones.
X=np.ones(len(x))
Y=np.ones(len(x))

# Populate the array with the data - both x and y are in the same txt file
for i in range(len(x)):
    X[i]=x[i][0]
    Y[i]=x[i][1]
#print(X,Y)

#print(X)
# Specifiy the bin width x5 ns (5ns is the value set in the Swabian) - so it multiplies the binwidth values set in the Swabian
binwidth=3
binx,biny = bin_data_given_bin_center(X,Y,nbin=int(len(X)/binwidth),opt='mean')
binx,_biny = bin_data_given_bin_center(X,Y,nbin=int(len(X)/binwidth),opt='std')

#plt.plot(X/1e3,Y)
#plt.errorbar(binx/1e3,biny,yerr=_biny)

# Normalization is done by taking the first 50/100 points of the data
norm1 = np.mean(Y[0:200])
#print(len(Y[0:50]))

# Normalize the data by the normalization factor calculated above
y1 = biny/ norm1

# Error on the Data
#y1_err = _biny/norm1        #error = std
y1_err = np.sqrt(biny)/norm1
#y1_err = np.sqrt(Y)/ norm1   # error sqrt(N_counts)
#y1_err = _biny/norm1

############
# Plotting #
############
# Set axes
#plt.axis([-1000, 1000, 0, 2])
#plt.axis([-600, 600, 0,2 ])

#print(binx)
#Plot the data
x1 = binx - 5.5e4
x1 = x1/1e6     # /1e6 to go from ps to us

################
# Plot g2 Data #
################

#plt.plot(x1, y1, marker='o', color='b')
plt.errorbar(x1, y1, yerr=y1_err, marker='.', markersize=10, color='royalblue', alpha=0.6, linestyle='None')
# print np.argmin(a)

#Custumize Plot
#plt.legend()
#plt.axis([-0.6, 0.6, 0,2 ])
plt.xlabel('Time ($\mu$s)', fontsize=16)
plt.ylabel(r'Normalized g$^{(2)}$ ($\tau $)', fontsize=16)
plt.rc('font', family='sans-serif')
plt.rc('font', serif='Helvetica')
plt.xticks(fontsize = 16)
plt.yticks(fontsize = 16)
plt.axhline(y=0.5, color='r', linestyle='--') #Plot a linea at 0.5
plt.title('Binning = ' + str(binwidth*5) + '_nsec')

###########
# Fitting #
###########
# # #Single Exponential - 2 Levels model
# # ## Ref. - https://arxiv.org/ftp/arxiv/papers/1607/1607.05348.pdf
# a = 0.5
# t1 = 0.6
# def _g2(t, a, t1 ):
#     #return 1 + a*np.exp(-(2*abs(t))/t1)
#     return 1 + a * np.exp(-abs(t)/t1)   # Single Exponential
# popt_g2, pcov_g2 = sp.optimize.curve_fit(_g2, x1, y1, p0=[a,t1])

###Double Exponential - Three-Level Model
###Eq. Reference Photon emission correlation spectroscopy as an analytical tool for quantum defects
###ref.2 - https://arxiv.org/ftp/arxiv/papers/1707/1707.04065.pdf
###Define your parameters - Initial Guess
# a = -0.03       # 4.60307898e-01
# t1 = 0.04       # short decay time
# t2 = 2 #0.5    # long decay time

# def _g2(t, a, t1, t2):
#     return 1 - (1+a)*np.exp(-abs(t)/t1) + a*np.exp(-abs(t)/t2)
#popt_g2, pcov_g2 = sp.optimize.curve_fit(_g2, x1, y1, p0=[a,t1,t2])

### Awschalom's Paper
a = 0.02 #100#5#0.03       # 4.60307898e-01
b = 0.25 #4# 0.01
t1 = 0.07       # short decay time
t2 = 2 #0.5    # long decay time

def _g2(t, a, b, t1,t2):
    return 1 - a*np.exp(-abs(t)/t1) + b*np.exp(-abs(t)/t2)
    #return 1 - a*np.exp(-abs(t)/t1) + b*np.exp(-t/t2)

popt_g2, pcov_g2 = sp.optimize.curve_fit(_g2, x1, y1, p0=[a,t1,t2,b])

###plt.plot(x1,_g2(x1,*popt_g2),'-',color='orange', label='Fitting') #+str(popt_lorentzian))
###plt.errorbar(x1,_g2(x1,*popt_g2), yerr=y1_err, fmt='-', color='orange', label='g2')

# Fitting Error
perr_g2 = np.sqrt(np.diag(pcov_g2))
#print(pcov_g2)
#print(perr_g2)
#print(popt_g2)

# Optimal Parameters rounded to 2 digits precision
a_opt = round(popt_g2[0],3)
t1_opt = round(popt_g2[1],3)
t2_opt = round(popt_g2[2],3)
b_opt = round(popt_g2[3],3)

# Error of the  Parameters rounded to 2 digits precision
a_opt_error = round(perr_g2[0],3)
t1_opt_error = round(perr_g2[1],3)
t2_opt_error = round(perr_g2[2],3)
b_opt_error = round(popt_g2[3],3)

print('Fitting Parameters:')
print('a =', a_opt, '+-', a_opt_error )
print('t1 =', t1_opt, '+-',t1_opt_error,'us')
print ('t2 =', t2_opt, '+-',t2_opt_error,'us')
print ('b =', b_opt, '+-',b_opt_error,'us')
print('The g2 minimum is', min(y1))
### Print the error of the minimum
print(x1[103],y1[103],y1_err[103])
#print('The fitting minimum is', min(y1))

# Initial Guess - Plot a function to match the data
# x_line=[]
# x_line=np.arange(min(x1),max(x1),1)
# y_line=_g2(x_line,1.02, 25.92,1)
# plt.plot(x_line,y_line)

# Confidence Interval
#https://stackoverflow.com/questions/24633664/confidence-interval-for-exponential-curve-fit

# Display
#plt.fill_between(x1, (_g2(x1,*popt_g2)-perr_g2), (_g2(x1,*popt_g2)+perr_g2), color='blue', alpha=0.1)
#plt.text(-1, 0.25, f"Tau = {t1*1000} +- {t1_opt_error} ms ", fontsize=12) ## Text inside the plot
plt.plot(x1,_g2(x1,*popt_g2),'-',color='orange') #label='Fit. func: $f(x) = %.f * e^{2*abs(t)/%.3f}' % (a_opt,t1_opt)
params_text = 'Fitting Parameters: \n a =' + str(a_opt) + '+-'+ str(a_opt_error) + \
                                  '\n t1 ='+ str(t1_opt) + '+-' + str(t1_opt_error) + ' us' \
                                  '\n t2 ='+ str(t2_opt) + '+-' + str(t2_opt_error) + ' us' \
                                   '\n b ='+ str(b_opt) + '+-' + str(b_opt_error) + ' us'
plt.text(1, 1.85, params_text, fontsize=14) ## text outside the plot
plt.legend(loc=1, prop={'size': 10})

plt.show()

#Save Graphs
#plt.savefig(fname=directory + 'g2_DLucky_2b_Laser_Stable_234_42457THz_2_5uW_Bin_5ns_2022-05-17_09-23-04_Binning_15ns.png')


#To do
## add error to the data --> sqrt(N)
## workout the unit of measure of tau
## subtract the dark counts
## add 95% confidence interval

# ###########################
# ### Fittings ####
# # ###########################


## Define the function you want to use for the fitting
# From Awschalom's Paper: g(2)(τ) = 1 − ae−τ/τ1 + be−τ/τ2, where a and b are
# amplitude parameters and τ1 and τ2 are the short and long decay times, respectively.
# The short decay time τ1 at 0.07(2) μs is reduced from the 0.17 μs optical
# lifetime (identical to ensemble) by optical Rabi driving.

# def _g2(a, b, t, t1, t2):
#     return 0.266 + 1 - a*np.exp(-t/t1) + b*np.exp(-t/t2)


## We then feed this function into a scipy function, along with our x- and y-axis data,
## and our guesses for the function fitting parameters:
## See https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html
#popt_g2, pcov_g2 = sp.optimize.curve_fit(_g2, x1, y1, p0=[a,b,t1,t2])


### Error
# # perr_gauss = np.sqrt(np.diag(pcov_gauss))
# #
# # pars_gauss = popt_gauss[0:3]
# # FWHM_gauss = 2*np.sqrt(2*np.log(2))*pars_gauss[2]
# #

############################
## Plot the data & Labels ##
############################
#plt.plot(x_values[0:-2], y_values[0:-2],'o', color='blue', label='Data - Sigma_M_QWP_37')
#plt.plot(x_values, y_values,'o', color='blue', label='Data - S-_QWP_37')
#plt.plot(x_values,_1gaussian(x_values,*popt_gauss),'o',color='orange', label='Gaussian fit')


#plt.plot(popt_gauss, pcov_gauss,'--')

## Save the Data
#plt.savefig(fname=directory + 'Fitting_D_Lucky_2b_0_05uW_Bg_Subtracted')
#

